/************************************************/
/*						*/
/*      Copyright:				*/
/*		 Jean-Marc Pigeon <jmp@safe.ca>	*/
/*	Distributed under the Gnu Public	*/
/*	License, see the License file in this	*/
/*	package.				*/
/*						*/
/*	CLEMENT					*/
/*	Its purpose is to handle SMTP protocole	*/
/*	on SMTP server premises and crunch data	*/
/*	and get clearance from the main Mentor	*/
/*	server to pass the E-mail to local	*/
/*	smtp server (sendmail, postfixe, etc)	*/
/*						*/
/************************************************/
#include	<syslog.h>
#include	<string.h>
#include	<errno.h>
#include	<signal.h>
#include	<stdio.h>
#include	<stdlib.h>
#include	<unistd.h>
#include	"modtra.h"
#include	"modchk.h"
#include	"modsnd.h"
#include	"modtrp.h"
#include	"lvleml.h"
#include	"lvlspm.h"
#include	"lvluid.h"
#include	"lvlchl.h"
#include	"gesclm.h"
#include	"unilck.h"
#include	"unieml.h"
#include	"unidns.h"
#include	"uninet.h"
#include	"subrou.h"
#include	"dbgmem.h"

typedef struct {	/*adv dispatcher struct		*/
	char *checker;	/*checker name			*/
	pid_t pid;	/*sending channel		*/
	char *msg;	/*message to send		*/
	}DISTYP;

static char curpath[300];
/*

*/
/************************************************/
/*						*/
/*	Procedure to display the program usage	*/
/*						*/
/************************************************/
static void usage()

{
(void) fprintf(stderr,"%s Version %s\n",appname,rou_getvers());
(void) fprintf(stderr,"usage:  %s "
		      "[-b base] "
		      "[-d debug] "
		      "[-f ] "
		      "[-h ] "
		      "[-v ] "
		      "\n",appname);
(void) fprintf(stderr,"\t\t-b base      : application working directory\n");
(void) fprintf(stderr,"\t\t-d debug     : debug level\n");
(void) fprintf(stderr,"\t\t-f           : foreground\n");
(void) fprintf(stderr,"\t\t-h           : display usage\n");
(void) fprintf(stderr,"\t\t-v           : verbose debug\n");
(void) exit(-1);
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to handle one communication	*/
/*	channel with a remote client. return	*/
/*	the foked process pid.			*/
/*						*/
/************************************************/
pid_t action(TRATYP *track,LSTTYP *channel,int num)

{
#define ACTCYC	"%s now %s [%s:%d] (%02d/%02d/%04d) (pid='%05d')"

pid_t child;

switch((child=rou_fork(false))) { 
  case  -1	:
    (void) alert(0,"%s:action Unable to start action process "
                   "(error=<%s>)",appname,strerror(errno));
    break;
  case   0	:	
      {
      int proceed;
      u_int cycle;
  
      (void) setjournal(WJRL,d_web,appname);
      memleak=DMTRACE;

      proceed=true;
      cycle=0;
      (void) journalling(true,APJRL,ACTCYC,"Modtra","Starting",
				    channel->ipnum,
		         	    channel->port,num+1,
				    channel->iteration,cycle,getpid());
      while (proceed==true) {
        int gotone;;

	gotone=0;
        (void) settitle("%s waiting on [%s:%d] (%02d/%02d/%04d)",
		         appname,channel->ipnum,
		         channel->port,num+1,channel->iteration,cycle);
        gotone=tra_tracking(track,channel);
        if ((quickout==false)&&(sigreload==false)) {
	  TIMESPEC timeout;

          timeout.tv_sec=0;
          timeout.tv_nsec=50000;
          while (ipc_waitpost(track->boxe,&timeout)==true) {
	    char *caller;
	    char *rqst;

	    if ((rqst=ipc_getpost(track->boxe,&caller))!=(char *)0) {
	      (void) alert(0,"action, Droping caller->msg=<%s->%s>",
			      caller,rqst);
	      caller=rou_freestr(caller);
  	      rqst=rou_freestr(rqst);
	      }
	    }
	  /*to make sure to have a new uniqID next time			*/
	  (void) nanosleep(&timeout,&timeout);
	  }
	else {
	  proceed=false;
	  }
        if ((gotone!=0)&&(proceed==true)) {
	  cycle++;
	  if (debug>0) {
            (void) journalling(true,APJRL,ACTCYC,"Modtra","cycling",
				          channel->ipnum,
		         	          channel->port,num+1,
				          channel->iteration,cycle,getpid());
	    }
	  }
        }
      (void) journalling(true,APJRL,ACTCYC,"Modtra","exiting",
				    channel->ipnum,
		         	    channel->port,num+1,
				    channel->iteration,cycle,getpid());
      /*Call to check any memory leak if memleak set to 'true',	*/
      /*request done in debug mode only.			*/
      (void) dbg_dumpmem(appname);
      memleak=DMFALSE;
      (void) closejournals();
      }
    (void) exit(0);
    break;
  default	:
    break;
  }
return child;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to load the config definition	*/
/*	and return a data structure		*/
/*						*/
/************************************************/
static CHLTYP *prepmng(CHLTYP *mng,BOXTYP *boxe,TRATYP *track)

{
CNFTYP *cnf;

if ((cnf=cnf_getconfig())!=(CNFTYP *)0) {
  int i;

  if (mng==(CHLTYP *)0) {
    mng=(CHLTYP *)calloc(1,sizeof(CHLTYP));
    mng->boxe=boxe;
    }
  mng->look=dev_getlst(mng->look,pro_smtp,cnf->bindlst);
  for (i=0;i<3;i++) {
    char **mout;
    char **minp;

    mout=(char **)0;
    minp=(char **)0;
    switch (i) {
      case 0	:
        mout=&(mng->chkidmod);
	minp=&(cnf->chkidmod);
	break;
      case 1	:
        mout=&(mng->rmponder);
	minp=&(cnf->rmponder);
	break;
      case 2	:
        mout=&(mng->spmdsrv);
	minp=&(cnf->spmdsrv);
	break;
      }
    *mout=rou_freestr(*mout); 
    if (*minp!=(char *)0)
      *mout=strdup(*minp);
    }
  (void) dev_mullisten(mng->look);
  mng->frelais=cnf->frelais;
  cnf->frelais=(char **)0;
  mng->relais=cnf->relais;
  cnf->relais=(IPTYP **)0;
  mng->auxs=cnf->auxs;
  cnf->auxs=(AUXTYP **)0;
  mng->lastload=cnf->lastload;
  track->dfs=eml_freeonedfs(track->dfs);
  track->dfs=eml_mkdf("default",cnf->spamlim,
		      cnf->qualoc,cnf->quahead,cnf->quaextra);
  track->skrlst=cnf_freeskrlst(track->skrlst);
  track->skrlst=cnf_getskrlst(mng->auxs);
  track->chkidsrv=rou_freestr(track->chkidsrv);
  track->chkidsrv=strdup(cnf->chkidsrv);
  track->dfltcode=strdup(cnf->dfltcode);
  track->chkidmod=rou_freestr(track->chkidmod);
  track->chkidmod=strdup(cnf->chkidmod);
  cnf=cnf_freeconfig(cnf);
  }
return mng;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to start process to listen	*/
/*	on ALL listed interface			*/
/*						*/
/************************************************/
static void dolistening(TRATYP *track,LSTTYP **look)

{
if (look!=(LSTTYP **)0) {
  int i;

  (void) dev_mullisten(look);
  for (i=0;look[i]!=(LSTTYP *)0;i++) {
    int j;

    if (look[i]->handle<0)
      continue;
    for (j=0;j<look[i]->iteration;j++) {
      if (look[i]->process[j]==(pid_t)0) {
        look[i]->process[j]=action(track,look[i],j);
        continue;
        }
      if (kill(look[i]->process[j],SIGCHECK)<0)
        look[i]->process[j]=(pid_t)0;
      }
    }
  }
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to check child process and	*/
/*	make sure they are up and running.	*/
/*						*/
/************************************************/
static void checkchild(CHLTYP *mng)

{
#define OPDR	"clement.c:checkchild,"

static time_t lastcheck=(time_t)0;

time_t curtime;

(void) rou_nozombie();
curtime=rou_systime();
if ((lastcheck+2)<=curtime) {
  ECHTYP t;
  int doupdate;

  doupdate=false;
  for (t=chl_sender;t<chl_unknown;t++) {
    if (rou_checkprocess(mng->pid[t])==false) {
      mng->newchild=true;
      if (lastcheck!=(time_t)0)
        (void) alert(0,"%s detect '%s' interface missing",OPDR,alrt[t]);
      switch (t) {
	case chl_sender		:
  	  mng->pid[(int)t]=snd_manager(mng);
	  break;
	case chl_checker	:
  	  mng->pid[(int)t]=chk_manager(mng);
	  break;
	case chl_transponder	:
	  if (mng->rmponder!=(char *)0)
  	    mng->pid[(int)t]=trp_manager(mng->rmponder);
	  else	/*no transponder defined	*/
  	    mng->pid[(int)t]=(pid_t)1;
	  break;
	default			:
	  (void) alert(0,"%s CHLTYP '%d' not expected",OPDR,(int)t);
	  break;
        }
      doupdate=true;
      }
    }
  (void) chl_dumpchldinfo(mng);
  lastcheck=curtime;
  }
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to stop all managed process 	*/
/*	and free the data structure.		*/
/*						*/
/************************************************/
static CHLTYP *dropmng(CHLTYP *mng)

{
if (mng!=(CHLTYP *)0) {

  (void) dev_mulclose(mng->look);
  mng->look=dev_freelst(mng->look);
  (void) rou_childkill(mng->pid,NSRV,5);
  mng->rmponder=rou_freestr(mng->rmponder);
  mng->frelais=rou_freestrlst(mng->frelais);
  mng->spmdsrv=rou_freestr(mng->spmdsrv);
  mng->chkidmod=rou_freestr(mng->chkidmod);
  (void) free(mng);
  mng=(CHLTYP *)0;
  }
return mng;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to open binding channel and	*/
/*	handle requested comming from SMTP	*/
/*	client.					*/
/*						*/
/************************************************/
static void  cycle()

{
#define	MXBOX	1000	/*Max 1000 messages queued*/

BOXTYP *boxe;
CHLTYP *mng;
TRATYP *track;

boxe=ipc_getbox(MXBOX);
mng=(CHLTYP *)0;
track=(TRATYP *)calloc(1,sizeof(TRATYP));
track->boxe=boxe;
track->clam=clm_opentree(track->clam);
(void) settitle("%s main daemon -d%d -b%s",appname,debug,base);
while (quickout==false) {
  if ((mng=prepmng(mng,boxe,track))==(CHLTYP *)0) {
    (void) alert(0,"%s configuration file missing altogether (aborting)",
		    appname);
    break;
    }
  (void) spm_isspamdup(mng->spmdsrv);
  (void) dolistening(track,mng->look);
  (void) checkchild(mng);
  (void) ipc_cleanbox(mng->boxe);
  (void) sleep(5);
  }
track=tra_freetrack(track);
mng=dropmng(mng);
boxe=ipc_dropbox(boxe);
}
/*
^L
*/
/************************************************/
/*						*/
/*	Program Entry.				*/
/*						*/
/************************************************/
int main(int argc,char *argv[])

{
char c;
STDDBG *dbglog;

memleak=DMFALSE;
base=strdup(base);
(void) strncpy(curpath,argv[0],sizeof(curpath));
(void) dbg_setcurpath(argv[0]);
(void) openlog(appname,LOG_PID,0);
while ((c=getopt(argc,argv,"b:d:fhv"))!=EOF) {
  switch(c) {
    case   'b'  :
      (void) free(base);
      base=rou_mkbase(optarg);
      break;
    case   'd'  :
      debug=atoi(optarg);
      break;
    case   'f'  :
      foreground=true;
      break;
    case   'v'  :
      verbose=true;
      break;
    case   'h'  :
    default	:
      (void) usage();
      break;
    }
  }
if ((debug==0)||(verbose==true)) {
  char curvers[100];
  char msg[200];

  (void) snprintf(curvers,sizeof(curvers),"%s",rou_getvers());
  (void) strcpy(msg,"");
  switch (dns_chksafeversion(curvers,sizeof(curvers))) {
    case -1	:
       (void) snprintf(msg,sizeof(msg),"\nYour current %s version is <%s> "
				       "while the last released one is <%s>.\n"
				        "You should update your %s program",
				        appname,rou_getvers(),curvers,appname);
      break;
    case  0	:
      break;
    case +1	:
       (void) snprintf(msg,sizeof(msg),"\nYour current %s-%s is a BETA "
				       "version.\n"
				       "Please report any discrepancy or bugs "
				       "you may find to <clement@safe.ca>.\n",
				        appname,rou_getvers());
      break;
    }
  if (strlen(msg)>0) {
    (void) fprintf(stderr,"%s\n",msg);
    }
  }
(void) rou_chgid(appname);
if ((strlen(base)>0)&&(chdir(base)<0)) {
  (void) fprintf(stderr,"Unable to reach the base directory <%s>\n",base);
  return -1;
  }
(void) chl_dropchldinfo();
(void) divedivedive(foreground);
if (lck_locking(appname,LCK_LOCK,5)==true) {
  dbglog=eml_switchstd(appname);
  (void) setjournal(APJRL,d_log,appname);
  (void) journalling(true,APJRL,"%s-%s started",appname,rou_getvers());
  (void) setjournal(EVJRL,d_log,"events");
  (void) preptitle(argc,argv,environ);
  (void) rou_setalrm();
  (void) cycle();
  (void) cleantitle();
  dbglog=eml_closeswitchstd(dbglog);
  (void) journalling(true,APJRL,"%s-%s stopped",appname,rou_getvers());
  (void) closejournals();
  (void) lck_locking(appname,LCK_UNLOCK,1);
  }
(void) closelog();
(void) free(base);
return 0;
}
