/************************************************/
/*						*/
/*      Copyright:				*/
/*		 Jean-Marc Pigeon <jmp@safe.ca>	*/
/*	Distributed under the Gnu Public	*/
/*	License, see the License file in this	*/
/*	package.				*/
/*						*/
/*	LVLCHL:					*/
/*	Define procedure to manage child process*/
/*	number.					*/
/*						*/
/************************************************/
#include	<stdio.h>
#include	<string.h>
#include	<errno.h>
#include	<sys/stat.h>
#include	"subrou.h"
#include	"lvlchl.h"

/*child pid filename (within /etc/clement-2.1)	*/
#define	ENFANTS	"children"

#define	SENDER	"sender"
#define	CHECKER	"checker"
#define	TRPONDR	"transponder"

const char *alrt[]={SENDER,CHECKER,TRPONDR,(char *)0};
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to create the 'children list'	*/
/*	filenane.				*/
/*						*/
/************************************************/
static char *getchildname(char *filename,int taille)

{
return rou_mkname(filename,taille,d_etc,"",ENFANTS,"");
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to scan the child pid file and*/
/*	collect pid related to auxiliairy child.*/
/*						*/
/************************************************/
static time_t scanchild(char *childname,pid_t childs[])

{
static VOCTYP vocable[]={
		{chl_sender,SENDER,(const void *)0},
		{chl_checker,CHECKER,(const void *)0},
		{chl_transponder,TRPONDR,(const void *)0},
		{chl_unknown,(char *)0,(const void *)0}
		};

time_t curtime;
FILE *fichier;

curtime=(time_t)0;
if ((fichier=fopen(childname,"r"))!=(FILE *)0) {
  VOCTYP *voc;
  char line[300];

  curtime=rou_systime();
  while (rou_getstr(fichier,line,sizeof(line),'#')!=(char *)0) {
    if ((voc=rou_getvoca(vocable,line))!=(VOCTYP *)0) {
      pid_t auxpid;

      auxpid=(pid_t)0;
      (void) sscanf(line,"%d",&auxpid);
      switch (voc->code) {
	case chl_sender		:
	case chl_checker	:
	case chl_transponder	:
	  childs[(int)voc->code]=auxpid;
	  break;
	default			:
	  (void) alert(0,"lvlchl.c:scanchild, Unknown directive <%s>",line);
	  break;
	}
      }
    }
  (void) fclose(fichier);
  }
return curtime;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to remove any remaining 	*/
/*	'child list' file			*/
/*						*/
/************************************************/
void chl_dropchldinfo()

{
struct stat bufstat;
char childname[300];

(void) getchildname(childname,sizeof(childname));
if (stat(childname,&bufstat)==0) {
  (void) unlink(childname);
  }
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to dump les informations a	*/
/*	propos du pid des programmes clients	*/
/*	dans un fichier.			*/
/*						*/
/************************************************/
void chl_dumpchldinfo(CHLTYP *mng)

{
if (mng->newchild==true) {
  FILE *fichier;
  char oldname[300];
  char newname[300];

  (void) getchildname(oldname,sizeof(oldname));
  (void) strcpy(newname,oldname);
  (void) strcat(newname,".new");
  if ((fichier=fopen(newname,"w"))!=(FILE *)0) {
    ECHTYP t;

    for (t=chl_sender;t<chl_unknown;t++)
      (void) fprintf(fichier,"%s %05d\n",alrt[t],mng->pid[t]);
    (void) fclose(fichier);
    (void) unlink(oldname);
    (void) link(newname,oldname);
    (void) unlink(newname);
    mng->newchild=false;
    }
  else {
    (void) core_dump("%s, Unable to open file <%s> (error=<%s>)",
		     appname,newname,strerror(errno));
    }
  }
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to return the current		*/
/*	auxiliary process number, if the file	*/
/*	has not been updated since last scan	*/
/*	keep the submitted number		*/
/*						*/
/************************************************/
pid_t chl_getchlpid(ECHTYP aux)

{
static time_t lastscan=(time_t)0;
static pid_t childs[(int)chl_unknown];

pid_t child;
int done;
char childname[300];

child=(pid_t)0;
(void) getchildname(childname,sizeof(childname));
done=false;
while (done==false) {
  done=true;
  switch (rou_isnewfile(childname,lastscan)) {
    case  1	:	/*scanner le fichier	*/
      lastscan=scanchild(childname,childs);
			  /*no break		*/
    case  0	:	/*file was not changed	*/
      child=childs[aux];
      break;
    case -1	:	/*file is missing!	*/
      done=false;	/*wait 1 ms to get it	*/
      (void) alert(0,"lvlchl.c:chl_getchlpid, file <%s> is missing!",
		      childname);
      (void) sleep(5);
      break;
    }
  }
return child;
}
